Namespace ChiSquaredUsingVisualBasic
    Friend Class ChiSquaredProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin chi-squared goodness-of-fit demo " & vbCrLf)

            ' 1. calculate chi-squared stat
            Console.WriteLine("Goal is to determine if roulette wheel is fair " & vbCrLf)

            Console.WriteLine("Wheel was spun 380 times " & vbCrLf)

            Dim observed() As Integer = {192, 163, 25} ' sum = 380 (180, 180, 20)
            Console.WriteLine("Observed counts of red, black, green: ")
            ShowVector(observed)

            Dim probs() As Double = {18.0 / 38, 18.0 / 38, 2.0 / 38}
            Console.WriteLine("Probabilities if fair:")
            ShowVector(probs, 4)

            Dim expected() As Double = ExpectedFromProbs(probs, 380)
            Console.WriteLine("Expected counts if fair:")
            ShowVector(expected, 1)

            Dim chi As Double = ChiFromProbs(observed, probs) ' 3.66
            Console.WriteLine("Calculated chi-squared = " & chi.ToString("F2"))

            ' 2. calculate p-value
            Dim df As Integer = observed.Length - 1
            Dim pval As Double = ChiSquarePval(chi, 2)
            Console.WriteLine("The pval with df of " & df & " = " & pval.ToString("F4"))

            Console.WriteLine(vbCrLf & "The pval is approximate probability that, if wheel is fair,")
            Console.WriteLine("you'd see a chi-squared value as extreme as calculated")

            Console.WriteLine(vbCrLf & "End demo")
            Console.ReadLine()
        End Sub ' Main

        Public Shared Sub ShowVector(ByVal v() As Integer)
            For i = 0 To v.Length - 1
                Console.Write(v(i) & "  ")
            Next i
            Console.WriteLine(vbCrLf)
        End Sub

        Public Shared Sub ShowVector(ByVal v() As Double, ByVal dec As Integer)
            For i = 0 To v.Length - 1
                Console.Write(v(i).ToString("F" & dec) & "  ")
            Next i
            Console.WriteLine(vbCrLf)
        End Sub

        Public Shared Function ChiFromFreqs(ByVal observed() As Integer, ByVal expected() As Double) As Double
            Dim sum As Double = 0.0
            For i = 0 To observed.Length - 1
                sum += ((observed(i) - expected(i)) * (observed(i) - expected(i))) / expected(i)
            Next i
            Return sum
        End Function

        Public Shared Function ChiFromProbs(ByVal observed() As Integer, ByVal probs() As Double) As Double
            Dim n As Integer = observed.Length
            Dim sumObs As Integer = 0
            For i = 0 To n - 1
                sumObs += observed(i)
            Next i

            Dim expected() As Double = ExpectedFromProbs(probs, sumObs)
            Return ChiFromFreqs(observed, expected)
        End Function

        Public Shared Function ExpectedFromProbs(ByVal probs() As Double, ByVal N As Integer) As Double()
            Dim expected(probs.Length - 1) As Double
            For i = 0 To probs.Length - 1
                expected(i) = probs(i) * N
            Next i
            Return expected
        End Function

        Public Shared Function ChiSquarePval(ByVal x As Double, ByVal df As Integer) As Double
            ' x = a computed chi-square value. df = degrees of freedom.
            ' output = prob. the x value occurred by chance.
            ' so, for example, if result < 0.05 there is only a 5% chance
            ' that the x value ocurred by chance and therefore
            ' we conclude that the actual data which produced x is
            ' NOT the same as the expected data.
            ' this function can be used to create a ChiSquareTest proedure.
            ' ACM Algorithm 299 and update ACM TOMS June 1985.
            ' Uses custom Exp() function below.

            If x <= 0.0 OrElse df < 1 Then
                Throw New Exception("Bad arg in ChiSquarePval()")
            End If

            Dim a As Double = 0.0 ' 299 variable names
            Dim y As Double = 0.0
            Dim s As Double = 0.0
            Dim z As Double = 0.0
            Dim ee As Double = 0.0 ' change from e
            Dim c As Double

            Dim even As Boolean ' is df even?

            a = 0.5 * x
            If df Mod 2 = 0 Then
                even = True
            Else
                even = False
            End If

            If df > 1 Then ' ACM update remark (4)
                y = Exp(-a)
            End If

            If even = True Then
                s = y
            Else
                s = 2.0 * Gauss(-Math.Sqrt(x))
            End If

            If df > 2 Then
                x = 0.5 * (df - 1.0)
                If even = True Then
                    z = 1.0
                Else
                    z = 0.5
                End If
                If a > 40.0 Then ' ACM remark (5)
                    If even = True Then
                        ee = 0.0
                    Else ' log(sqrt(pi))
                        ee = 0.57236494292470008
                    End If
                    c = Math.Log(a) ' log base e
                    Do While z <= x
                        ee = Math.Log(z) + ee
                        s = s + Exp(c * z - a - ee) ' ACM update remark (6)
                        z = z + 1.0
                    Loop
                    Return s ' a > 40.0
                Else
                    If even = True Then
                        ee = 1.0
                    Else ' (1 / sqrt(pi))
                        ee = 0.56418958354775628 / Math.Sqrt(a)
                    End If
                    c = 0.0
                    Do While z <= x
                        ee = ee * (a / z) ' ACM update remark (7)
                        c = c + ee
                        z = z + 1.0
                    Loop
                    Return c * y + s
                End If ' df > 2
            Else
                Return s
            End If
        End Function ' ChiSquarePval()

        Private Shared Function Exp(ByVal x As Double) As Double ' ACM update remark (3)
            If x < -40.0 Then ' ACM update remark (8)
                Return 0.0
            Else
                Return Math.Exp(x)
            End If
        End Function

        Public Shared Function Gauss(ByVal z As Double) As Double
            ' input = z-value (-inf to +inf)
            ' output = p under Normal curve from -inf to z
            ' e.g., if z = 0.0, function returns 0.5000
            ' ACM Algorithm #209
            Dim y As Double ' 209 scratch variable
            Dim p As Double ' result. called ‘z’ in 209
            Dim w As Double ' 209 scratch variable

            If z = 0.0 Then
                p = 0.0
            Else
                y = Math.Abs(z) / 2
                If y >= 3.0 Then
                    p = 1.0
                ElseIf y < 1.0 Then
                    w = y * y
                    p = ((((((((0.000124818987 * w - 0.001075204047) * w + 0.005198775019) * w - 0.019198292004) * w + 0.059054035642) * w - 0.151968751364) * w + 0.319152932694) * w - 0.5319230073) * w + 0.797884560593) * y * 2.0
                Else
                    y = y - 2.0
                    p = (((((((((((((-0.000045255659 * y + 0.00015252929) * y - 0.000019538132) * y - 0.000676904986) * y + 0.001390604284) * y - 0.00079462082) * y - 0.002034254874) * y + 0.006549791214) * y - 0.010557625006) * y + 0.011630447319) * y - 0.009279453341) * y + 0.005353579108) * y - 0.002141268741) * y + 0.000535310849) * y + 0.999936657524
                End If
            End If

            If z > 0.0 Then
                Return (p + 1.0) / 2
            Else
                Return (1.0 - p) / 2
            End If
        End Function ' Gauss()
    End Class ' Program
End Namespace ' ns
